import duckdb
from core_pro.ultilities import make_sync_folder
from pathlib import Path
from datasets import Dataset
import numpy as np
import sys
sys.path.extend([str(Path.home() / 'PycharmProjects/item_matching')])

from src.item_matching.model.model import Model


# path
cluster = 'FMCG'
path = make_sync_folder("dataset/item_matching")

file = path / f'data_sample_{cluster}_clean.parquet'
file_inner = path / 'inner.parquet'

# data loading
query = f"""
with base as (
select * exclude(level1_global_be_category, level2_global_be_category, level3_global_be_category)
, level1_global_be_category || '__' || level2_global_be_category || '__' || level3_global_be_category category
from read_parquet('{file}')
)
select *
from base
"""
df = (
    duckdb.sql(query).pl()
    .unique(['item_id'])
)
dataset_chunk = Dataset.from_polars(df)

model = Model()
model.get_text_model()
embeddings = model.process_text(dataset_chunk["item_name_clean"])
np.save(path / "embeddings.npy", embeddings)



from pathlib import Path
from pydantic import BaseModel, Field, computed_field
import polars as pl
import duckdb
import re
from time import perf_counter
from core_pro.ultilities import rm_all_folder, make_dir
from .pipeline.build_index_and_query import BuildIndexAndQuery, ConfigQuery
from .pipeline.data_loading import DataEmbedding, ConfigEmbedding
from rich import print


class MatchInput(BaseModel):
    ROOT_PATH: Path = Field(default=None)
    PATH_Q: Path = Field(default=None)
    PATH_DB: Path = Field(default=None)
    PATH_INNER: Path = Field(default=None)
    INNER: bool = Field(default=False)
    MATCH_BY: str = Field(default="text")  # image
    COL_CATEGORY: str = Field(default="")
    SHARD_SIZE: int = Field(default=1_500_000)
    QUERY_SIZE: int = Field(default=50_000)
    TOP_K: int = Field(default=10)
    EXPLODE: bool = Field(default=True)

    @computed_field
    @property
    def path_result(self) -> Path:
        path_result = self.ROOT_PATH / f"result_match_{self.MATCH_BY}"
        make_dir(path_result)
        return path_result


class PipelineMatch:
    def __init__(self, record: MatchInput):
        # path
        self.PATH_Q = record.PATH_Q
        self.PATH_DB = record.PATH_DB
        self.PATH_INNER = record.PATH_INNER
        self.ROOT_PATH = record.ROOT_PATH
        self.PATH_RESULT = record.path_result

        # config

        self.COL_CATEGORY = record.COL_CATEGORY
        self.MATCH_BY = record.MATCH_BY
        self.SHARD_SIZE = record.SHARD_SIZE
        self.QUERY_SIZE = record.QUERY_SIZE
        self.TOP_K = record.TOP_K
        self.INNER = record.INNER
        self.EXPLODE = record.EXPLODE

        self.lst_category = self._category_chunking()

    def _category_chunking(self) -> list:
        """
        Read by duckdb to perform lazy load
        """
        query = f"""
        select distinct {{1}}{self.COL_CATEGORY} as category 
        from read_parquet('{{0}}') 
        where {{1}}{self.COL_CATEGORY} is not null
        """
        query = (
            query.format(self.PATH_INNER, "")
            if self.INNER
            else query.format(self.PATH_Q, "q_")
        )
        lst_category = duckdb.sql(query).pl()["category"].to_list()
        return sorted(lst_category)

    def _load_data(self, cat: str, mode: str, file: Path):
        """
        Read by polars to prevent special characters in writing query
        """
        filter_ = (
            pl.col(f"{self.COL_CATEGORY}") == cat
            if self.INNER
            else pl.col(f"{mode}_{self.COL_CATEGORY}") == cat
        )
        return pl.read_parquet(file).filter(filter_)

    def run(self):
        # run
        start = perf_counter()
        for idx, cat in enumerate(self.lst_category):
            # logging
            cat_log = f"[dark_orange]{cat}[/]"
            batch_log = f"{idx}/{len(self.lst_category) - 1}"

            # check file exists
            file_result_final = self.PATH_RESULT / f"{cat}.parquet"
            if file_result_final.exists():
                print(f"[PIPELINE] File {cat_log} already exists")
                continue

            # chunk checking
            print("*" * 50)
            print(
                f"🐋 [PIPELINE MATCH BY {self.MATCH_BY}] 🐋 \n"
                f"-> Category: {cat_log} {batch_log}"
            )
            if not self.INNER:
                chunk_db = self._load_data(cat=cat, mode="db", file=self.PATH_DB)
                chunk_q = self._load_data(cat=cat, mode="q", file=self.PATH_Q)
                print(
                    f"-> Database shape {chunk_db.shape}, Query shape {chunk_q.shape}"
                )

                if chunk_q.shape[0] < 2 or chunk_db.shape[0] < 2:
                    print(f"[PIPELINE] Database/Query have not enough data")
                    continue

                # embeddings
                config = ConfigEmbedding(
                    ROOT_PATH=self.ROOT_PATH,
                    MODE="db",
                    SHARD_SIZE=self.SHARD_SIZE,
                    MATCH_BY=self.MATCH_BY,
                )
                DataEmbedding(config_input=config).load(data=chunk_db)

                config = ConfigEmbedding(
                    ROOT_PATH=self.ROOT_PATH,
                    MODE="q",
                    SHARD_SIZE=self.SHARD_SIZE,
                    MATCH_BY=self.MATCH_BY,
                )
                DataEmbedding(config_input=config).load(data=chunk_q)

            else:
                chunk_df = self._load_data(cat=cat, mode="db", file=self.PATH_INNER)
                print(f"-> Inner Data shape {chunk_df.shape}")

                if chunk_df.shape[0] < 2:
                    print(f"[PIPELINE] Database/Query have no data")
                    continue

                # embeddings
                config = ConfigEmbedding(
                    ROOT_PATH=self.ROOT_PATH,
                    MODE="",
                    SHARD_SIZE=self.SHARD_SIZE,
                    MATCH_BY=self.MATCH_BY,
                )
                DataEmbedding(config_input=config).load(data=chunk_df)

            # index and query
            config = ConfigQuery(
                ROOT_PATH=self.ROOT_PATH,
                QUERY_SIZE=self.QUERY_SIZE,
                MATCH_BY=self.MATCH_BY,
                TOP_K=self.TOP_K,
            )
            cat = re.sub("/", "", cat)  # special characters
            build = BuildIndexAndQuery(
                config=config,
                file_export_name=cat,
                inner=self.INNER,
                explode=self.EXPLODE,
            )
            build.build()
            build.query()

            folder_list = [
                "index",
                "result",
                "db_array",
                "db_ds",
                "q_array",
                "q_ds",
                "array",
                "ds",
            ]
            for name in folder_list:
                rm_all_folder(self.ROOT_PATH / name)

        time_perf = perf_counter() - start
        print(
            f"🐋 [PIPELINE MATCH BY {self.MATCH_BY}] 🐋 \n"
            f"-> Your files are ready, please find here: {self.PATH_RESULT}"
        )
        return {f"time_perf_{self.MATCH_BY}": time_perf}



from pathlib import Path
import duckdb


path = Path("/Users/kevinkhang/Downloads/Data")

file_brand = path / "brand/brand.parquet"
file_data = path / "Item_Matching_Test/clean.parquet"

query = f"""
select level1_global_be_category
 , array_agg(distinct lower(final_brand)) final_brand
 from read_parquet('{file_brand}')
 group by 1
"""
df_l1_brand = duckdb.sql(query).pl()

query = f"""
select final_brand
 , array_agg(distinct lower(level1_global_be_category)) level1_global_be_category
 , len(array_agg(distinct lower(level1_global_be_category))) len
 from read_parquet('{file_brand}')
 group by 1
"""
df_brand = duckdb.sql(query).pl()

# query = f"""
# with brand as (
#     select level1_global_be_category
#      , lower(final_brand) final_brand
#      from read_parquet('{file_brand}')
# )
# select i.*
# , b.final_brand
# from read_parquet('{file_data}') i
# left join brand as b
# on i.level1_global_be_category = b.level1_global_be_category
# """
# df = duckdb.sql(query).pl()
# print(df.shape)



from pathlib import Path
import polars as pl
from rich import print
from core_pro import ImageDownloader
from core_pro.ultilities import make_dir
from core_eda import TextEDA


class PipelineImage:
    def __init__(
        self,
        path_image: Path,
        mode: str = "",
        col_img_download: str = "image_url",
        col_text: str = "item_name",
    ):
        # path
        self.path_image = path_image
        self.mode = mode
        self.folder_image = self.path_image / f"img_{self.mode}"

        # config
        self.col_img_download = col_img_download
        self.col_text = col_text
        make_dir(self.path_image)

        print(f"[Image Processing] {mode}")

    def load_images(self) -> pl.DataFrame:
        lst = [(int(i.stem), str(i), i.exists()) for i in self.folder_image.glob("*/*.jpg")]
        df = (
            pl.DataFrame(lst, orient="row", schema=["index", "file_path", "exists"])
            .with_columns(pl.col("index").cast(pl.UInt32))
        )
        if df.shape[0] == 0:
            print(f"-> Images Errors {self.mode}: {df.shape}")
        else:
            print(f"-> Load images from folder {self.mode}: {df.shape}")
        return df

    def run(
        self,
        data,
        download: bool = False,
        num_processes: int = 4,
        num_workers: int = 16,
    ):
        # load data
        data = data.with_row_index()
        run = data[self.col_img_download].to_list()
        print(f"-> Base data {self.mode}: {data.shape}")

        # download
        if download:
            downloader = ImageDownloader(
                output_dir=self.folder_image,
                num_processes=num_processes,
                threads_per_process=num_workers,
                resize_size=(224, 224),
                batch_size=1000,
                images_per_folder=1000,
            )
            downloader.download_images(run)

        # load data image
        data_img = self.load_images()

        # join
        data = (
            data.pipe(TextEDA.clean_text_pipeline_polars, col=self.col_text)
            .join(data_img, on="index", how="left")
            .filter(pl.col("exists"))
        )
        if self.mode != "":
            data = data.select(pl.all().name.prefix(f"{self.mode}_"))

        print(f"-> Merge and clean images data {self.mode}: {data.shape}")
        return data, data_img



import polars as pl
import duckdb
from pathlib import Path
from pydantic import BaseModel, Field, computed_field
from tqdm import tqdm
from rich import print
from core_pro.ultilities import make_dir
from ..func.post_processing import data_explode_list


class ReRankConfig(BaseModel):
    ROOTPATH: Path = Field(default=None)
    EXPLODE: bool = Field(default=False)
    col_text: str = Field(default="item_name")

    @computed_field
    @property
    def path_result(self) -> Path:
        path_result = self.ROOTPATH / "result_match_rerank"
        make_dir(path_result)
        return path_result

    @computed_field
    @property
    def path_text(self) -> Path:
        return self.ROOTPATH / "result_match_text"

    @computed_field
    @property
    def path_image(self) -> Path:
        return self.ROOTPATH / "result_match_image"

    @computed_field
    @property
    def file_text(self) -> list:
        return [*self.path_text.glob("*.parquet")]

    @computed_field
    @property
    def file_image(self) -> list:
        return [*self.path_image.glob("*.parquet")]


class ReRank:
    def __init__(self, record: ReRankConfig):
        # all category
        self.all_category = set(
            [i.stem for i in record.file_text] + [i.stem for i in record.file_image]
        )

        # config
        self.EXPLODE = record.EXPLODE

        # path
        self.path_result = record.path_result
        self.path_image = record.path_image
        self.path_text = record.path_text

        # file
        self.file_text = record.file_text
        self.file_image = record.file_image

        # col
        self.col_text = record.col_text

    def _data_check(self, category: str):
        paths = {"text": self.path_text, "image": self.path_image}
        df_dict = {}
        for name, path in paths.items():
            df = pl.read_parquet(path / f"{category}.parquet")
            if self.EXPLODE:
                df = data_explode_list(df)
            df_dict[name] = df
        return df_dict

    def rerank_score(self, data_text, data_image):
        # query
        query = f"""
            -- combine text & image
            with base as (
                select q_index
                , q_{self.col_text}
                , db_index
                , db_{self.col_text}
                , 'score_text' match_type
                , score_text_embed score
                from data_text
                union all
                select q_index
                , q_{self.col_text}
                , db_index
                , db_{self.col_text}
                , 'score_image' match_type
                , score_image_embed score
                from data_image
            ) 
            -- pivot scores into 2 cols
            , pivot_tab as (
                PIVOT base
                ON match_type
                USING sum(score)
                GROUP BY q_index, q_{self.col_text}, db_index, db_{self.col_text}
            )
            -- calculate mean, max rerank
            , cal_tab as (
                select *
                , (score_text + score_image) / 2 as score_mean
                from pivot_tab
            )
            -- rank matches
            select * exclude(score_mean)
            , coalesce(score_mean, score_text, score_image) score_mean
            from cal_tab
            order by q_index, coalesce(score_mean, score_text, score_image) desc
            """
        return duckdb.sql(query).pl()

    def run(self):
        total_cat = len(self.all_category)
        for cat in tqdm(self.all_category, total=total_cat, desc="ReRanking"):
            file_name = self.path_result / f"{cat}.parquet"
            if file_name.exists():
                continue
            df_dict = self._data_check(cat)
            df = self.rerank_score(
                data_text=df_dict["text"], data_image=df_dict["image"]
            )
            df.write_parquet(file_name)
        print(f"[RERANK]: Done {total_cat} categories")



from core_pro import DataPipeLine
from core_pro.ultilities import make_sync_folder


query = f"""
with base as (
    select distinct
        level1_global_be_category
        ,final_brand
    from mp_mall.dim_brand_item_brandmapping__vn_s0_live
    where
        grass_date = current_date - interval '1' day
        and global_brand != 'NoBrand'
        and global_brand is not null
        and status = 1
)
select
    *
from base
"""
path = make_sync_folder('brand')
df = DataPipeLine(query).run_presto_to_df(save_path=path / 'brand.parquet')



from .main_match import PipelineMatch, MatchInput
from .func.function_image import PipelineImage
from .func.function_text import PipelineText
from .func.post_processing import PostProcessing
from .pipeline.rerank import ReRank, ReRankConfig



from core_pro import DataPipeLine
from core_pro.ultilities import make_sync_folder, update_df
import polars as pl


# download parquet
path = make_sync_folder("Item_Matching_Test")
sh = "1JQiN33Z8hTPd_ENBAoqToMKXGr2Qp5ZU0rjvFPHCTAI"

query = f"""
with cat as (
    select distinct
        level1_global
        ,cluster
    from vnbi_mkt.general__global_category_map
    where
        ingestion_timestamp in (
            select
                max(ingestion_timestamp)
            from vnbi_mkt.general__global_category_map
        )
)
select cluster
    ,o.level1_global_be_category
    ,o.level2_global_be_category
    ,o.level3_global_be_category
    ,count(distinct case when o.grass_date >= current_date - interval '30' day then o.item_id else null end) total_items_30d
    ,count(distinct case when o.grass_date >= current_date - interval '60' day then o.item_id else null end) total_items_60d
    ,count(distinct o.item_id) total_items_90d
from
    mp_order.dwd_order_item_all_ent_df__vn_s0_live o
    left join cat m on m.level1_global = o.level1_global_be_category
where
    o.grass_date >= current_date - interval '90' day
group by 1, 2, 3, 4
"""
save_path = path / "stats/stats_l3.parquet"
df = DataPipeLine(query).run_presto_to_df(save_path=save_path)

df = (
    pl.read_parquet(save_path)
    .with_columns(
        (pl.col("total_items_90d") / pl.col("total_items_90d").sum()).alias("pct_90d"),
        (pl.col("total_items_60d") / pl.col("total_items_30d")).alias("dif_60"),
        (pl.col("total_items_90d") / pl.col("total_items_30d")).alias("dif_90"),
    )
    .drop_nulls(subset=["cluster"])
    .sort(["level1_global_be_category"])
)
update_df(df, "category", sh, start="A1")

query = f"""
with cat as (
    select distinct
        level1_global
        ,cluster
    from vnbi_mkt.general__global_category_map
    where
        ingestion_timestamp in (
            select
                max(ingestion_timestamp)
            from vnbi_mkt.general__global_category_map
        )
)
select cluster
    ,o.level1_global_be_category
    ,o.level2_global_be_category
    ,o.level3_global_be_category
    ,o.level4_global_be_category
    ,o.level5_global_be_category
    ,count(distinct case when o.grass_date >= current_date - interval '30' day then o.item_id else null end) total_items_30d
    ,count(distinct case when o.grass_date >= current_date - interval '60' day then o.item_id else null end) total_items_60d
    ,count(distinct o.item_id) total_items_90d
from
    mp_order.dwd_order_item_all_ent_df__vn_s0_live o
    left join cat m on m.level1_global = o.level1_global_be_category
where
    o.grass_date >= current_date - interval '90' day
group by 1, 2, 3, 4, 5, 6
"""
save_path = path / "stats/stats_l5.parquet"
df = DataPipeLine(query).run_presto_to_df(save_path=save_path)

df = (
    pl.read_parquet(save_path)
    .with_columns(
        (pl.col("total_items_90d") / pl.col("total_items_90d").sum()).alias("pct_90d"),
        (pl.col("total_items_60d") / pl.col("total_items_30d")).alias("dif_60"),
        (pl.col("total_items_90d") / pl.col("total_items_30d")).alias("dif_90"),
    )
    .drop_nulls(subset=["cluster"])
    .sort(["level1_global_be_category"])
)
update_df(df, "category", sh, start="M1")



from time import perf_counter
from sentence_transformers import SentenceTransformer
from FlagEmbedding import BGEM3FlagModel
from notebooks.benchmark_database.data_load import load
import polars as pl
from collections import defaultdict
from pathlib import Path
import numpy as np


# init data
df, col, _ = load()


def jina(texts: list):
    start = perf_counter()
    model = SentenceTransformer("jinaai/jina-embeddings-v3", trust_remote_code=True)
    model.half()
    model.max_seq_length = 80
    task = 'text-matching'
    embeddings = model.encode(
        texts,
        task=task,
        prompt_name=task,
        show_progress_bar=True
    )
    estimated_time = perf_counter() - start
    file_embed = path_array / 'jina.npy'
    np.save(file_embed, embeddings)
    return estimated_time, embeddings.shape


def bge(texts: list):
    start = perf_counter()
    model = BGEM3FlagModel('BAAI/bge-m3', use_fp16=True)
    embeddings = model.encode(
        texts,
        batch_size=8,
        max_length=80,
        return_dense=True,
        return_sparse=False,
        return_colbert_vecs=False
    )['dense_vecs']
    estimated_time = perf_counter() - start
    file_embed = path_array / 'bge.npy'
    np.save(file_embed, embeddings)
    return estimated_time, embeddings.shape


path = Path('/media/kevin/data_4t/Test/benchmark_database')
path_array = path / 'array'

lst = df['q_item_name_clean'].to_list()
dict_ = defaultdict(list)
bench = {
    'jina': jina,
    'bge': bge,
}
for i, f in bench.items():
    time, size = f(lst)
    dict_['name'].append(i)
    dict_['time'].append(time)
    dict_['data_size'].append(size[0])
    dict_['embed_size'].append(size[1])

report = (
    pl.DataFrame(dict_)
    .with_columns((pl.col('data_size') / pl.col('time')).alias('per_sec'))
)
report.write_csv(path / 'model_benchmark.csv')



import duckdb
from pathlib import Path
from re import search
from core_pro.ultilities import make_dir


def data_explode_list(data):
    col_explode = [i for i in data.columns if search("db|score", i)]
    return data.explode(col_explode)


class PostProcessing:
    def __init__(self, data, verbose: bool = False):
        self.data = data
        self.verbose = verbose
        self.final_col = None
        self.query_add_url_duckdb = ""
        self.query_add_show_img = ""
        self.mode = ["db", "q"]

    @staticmethod
    def select_export_cols(data):
        q_col = sorted(
            filter(lambda x: "q" in x and "image_url" not in x, data.columns)
        )
        db_col = sorted(
            filter(lambda x: "db" in x and "image_url" not in x, data.columns)
        )
        score_col = sorted(filter(lambda x: "score" in x, data.columns))
        final_col = q_col + db_col + ["q_image_url", "db_image_url"] + score_col
        return final_col

    def query_add_url(self) -> str:
        for i in self.mode:
            if f"{i}_shop_id" in self.data.columns:
                self.query_add_url_duckdb += f", 'https://shopee.vn/product/' || {i}_shop_id || '/' || {i}_item_id {i}_url\n"
        return self.query_add_url_duckdb

    def query_add_show_image(self) -> str:
        for i in self.mode:
            if f"{i}_image_url" in self.data.columns:
                self.query_add_show_img += (
                    f""", '=IMAGE("' || {i}_image_url || '", 1)' {i}_image_url\n"""
                )
        return self.query_add_show_img

    def run(self):
        # init
        show_image = self.query_add_show_image()
        add_exclude = ""
        if len(show_image) != 0:
            add_exclude += "EXCLUDE (db_image_url, q_image_url)"
        data = self.data

        # query
        query = f"""
        select * {add_exclude}
        {show_image}
        {self.query_add_url()}
        from data
        """
        if self.verbose:
            print(query)
        else:
            df = duckdb.sql(query).pl()
            select_col = PostProcessing.select_export_cols(df)
            df = df.select(select_col)
        return df

    @staticmethod
    def save_parquet_to_csv(path: Path):
        path_export = path.parent / "result_export"
        make_dir(path_export)
        for f in [*path.glob("*.parquet")]:
            query = f"""COPY (SELECT * FROM read_parquet('{str(f)}')) TO '{path_export}/{f.stem}.csv' (HEADER, DELIMITER ',')"""
            duckdb.sql(query)
        return path_export



import torch
from PIL import Image
from torch.nn import functional as F
from transformers import AutoProcessor, SiglipVisionModel
from FlagEmbedding import BGEM3FlagModel
from accelerate.test_utils.testing import get_backend


class Model:
    def __init__(self, max_length: int = 50):
        self.device, _, _ = get_backend()
        self.max_length = max_length

        self.text_model = None
        self.img_processor = None
        self.img_model = None

    def get_img_model(self):
        """Get Image Model"""
        model_id = "google/siglip-base-patch16-224"
        self.img_processor = AutoProcessor.from_pretrained(model_id)

        config = {
            "pretrained_model_name_or_path": model_id,
            "torch_dtype": torch.bfloat16,
        }
        self.img_model = SiglipVisionModel.from_pretrained(**config).to(self.device)
        self.img_model = torch.compile(self.img_model)

    def get_text_model(self):
        model_id = "BAAI/bge-m3"
        self.text_model = BGEM3FlagModel(model_id, use_fp16=True)

    def process_text(self, list_text: list[str]):
        embeddings = self.text_model.encode(
            list_text,
            batch_size=512,
            max_length=self.max_length,
            return_dense=True,
            return_sparse=False,
            return_colbert_vecs=False,
        )["dense_vecs"]
        return embeddings

    def process_image(self, batch, col: str):
        images = [Image.open(i).convert("RGB") for i in batch[col]]
        inputs = self.img_processor(images=images, return_tensors="pt").to(self.device)
        with torch.inference_mode():
            outputs = self.img_model(**inputs)
        pooled_output = outputs.pooler_output
        return {"image_embed": pooled_output}

    @staticmethod
    def pp_normalize(batch, col: str = "image_embed") -> dict:
        norm_embed = F.normalize(batch[col], p=2, dim=1).cpu().numpy()
        return {col: norm_embed}



from pydantic import BaseModel, Field, computed_field
import polars as pl
from datasets import Dataset, concatenate_datasets
from pathlib import Path
import numpy as np
from rich import print
from core_pro.ultilities import create_batch_index, make_dir
from ..model.model import Model


class ConfigEmbedding(BaseModel):
    ROOT_PATH: Path = Field(default=None)
    SHARD_SIZE: int = Field(default=1_500_000)
    MODE: str = Field(default="")
    MATCH_BY: str = Field(default="text")

    @computed_field
    @property
    def col_embedding(self) -> str:
        dict_ = {"image": "image_embed"}
        return dict_.get(self.MATCH_BY, "text_embed")

    @computed_field
    @property
    def col_input(self) -> str:
        col = "file_path" if self.MODE == "" else f"{self.MODE}_file_path"
        default_val = (
            "item_name_clean" if self.MODE == "" else f"{self.MODE}_item_name_clean"
        )
        dict_ = {"image": col}
        return dict_.get(self.MATCH_BY, default_val)

    @computed_field
    @property
    def path_array(self) -> Path:
        folder = "array" if self.MODE == "" else f"{self.MODE}_array"
        return self.ROOT_PATH / folder

    @computed_field
    @property
    def path_ds(self) -> Path:
        folder = "ds" if self.MODE == "" else f"{self.MODE}_ds"
        return self.ROOT_PATH / folder


class DataEmbedding:
    def __init__(self, config_input: ConfigEmbedding):
        # Path
        self.MATCH_BY = config_input.MATCH_BY
        self.SHARD_SIZE = config_input.SHARD_SIZE
        self.col_embedding = config_input.col_embedding
        self.col_input = config_input.col_input
        self.path_array = config_input.path_array
        self.path_ds = config_input.path_ds

        # Init path
        make_dir(self.path_array)
        make_dir(self.path_ds)

    def load(self, data: pl.DataFrame):
        # Log total chunks
        run = create_batch_index(data.shape[0], self.SHARD_SIZE)
        num_chunks = len(run)

        # Model
        model = Model()
        if self.MATCH_BY == "text":
            model.get_text_model()
        elif self.MATCH_BY == "image":
            model.get_img_model()

        # Process and save each chunk
        for i, idx in run.items():
            # Check if exists:
            dataset_name = self.path_ds / f"{i}"
            array_name = self.path_array / f"{i}.npy"
            if dataset_name.exists():
                continue

            # Load Chunk
            start_idx, end_idx = idx[0], idx[-1]
            if start_idx == end_idx:  # prevent sample size is 1
                end_idx = None

            dataset_chunk = Dataset.from_polars(data[start_idx:end_idx])
            print(
                f"[DataEmbedding] Shard [{i}/{num_chunks - 1}]: start {start_idx:,.0f} end {end_idx:,.0f}"
            )

            # Process dataset
            if self.MATCH_BY == "text":
                embeddings = model.process_text(dataset_chunk[self.col_input])
                dset_embed = Dataset.from_dict({self.col_embedding: embeddings})
                dataset_chunk = concatenate_datasets(
                    [dataset_chunk, dset_embed], axis=1
                )
            else:
                dataset_chunk = dataset_chunk.map(
                    model.process_image,
                    batch_size=512,
                    batched=True,
                    fn_kwargs={"col": self.col_input},
                )

            # Normalize
            dataset_chunk.set_format(
                type="torch", columns=[self.col_embedding], output_all_columns=True
            )
            if self.MATCH_BY == "image":
                dataset_chunk = dataset_chunk.map(
                    Model.pp_normalize,
                    batched=True,
                    fn_kwargs={"col": self.col_embedding},
                )
            dataset_chunk.set_format(
                type="numpy", columns=[self.col_embedding], output_all_columns=True
            )

            # Save chunk
            np.save(array_name, dataset_chunk[self.col_embedding])
            dataset_chunk.save_to_disk(str(dataset_name))



from rich import print
import duckdb
import polars as pl
from core_pro import DataPipeLine
from core_pro.ultilities import make_sync_folder
from core_eda import EDA_Dataframe
from pathlib import Path
import sys

sys.path.extend([str(Path.home() / "PycharmProjects/item_matching")])

from src.item_matching import PipelineImage


# download parquet
path = make_sync_folder("dataset/item_matching")
path_sql = Path.home() / "PycharmProjects/item_matching/notebooks/data/data.sql"
path_image = path / "img"

# for c in ['ELHA', 'Lifestyle', 'Fashion', 'FMCG']:
for c in ["FMCG"]:
    file = path / f"data_sample_{c}.parquet"
    if not file.exists():
        print(f"Download {file.stem}")
        sql = open(str(path_sql)).read().format(c)
        df = DataPipeLine(sql).run_presto_to_df(save_path=file)
    else:
        df = pl.read_parquet(file)
        print(f"File {file.stem} exists")

    # download images
    query = f"""
    select *
    ,concat('http://f.shopee.vn/file/', UNNEST(array_slice(string_split(images, ','), 1, 1))) image_url
    from read_parquet('{str(path / f'{file.stem}.parquet')}')
    order by item_id, images
    """
    df = duckdb.sql(query).pl().unique(["item_id"])
    EDA_Dataframe(df, ["item_id"]).check_duplicate()

    df, df_img = PipelineImage(
        path_image=path_image, mode="", col_img_download="image_url"
    ).run(df, download=True, num_workers=4, num_processes=4)
    df.write_parquet(path / f"data_sample_{c}_clean.parquet")



from pathlib import Path
import polars as pl
from time import perf_counter
from autofaiss import build_index
from datasets import concatenate_datasets, load_from_disk
from pydantic import BaseModel, Field, computed_field
import numpy as np
from rich import print
from core_pro.ultilities import create_batch_index, make_dir
from ..func.post_processing import data_explode_list


class ConfigQuery(BaseModel):
    ROOT_PATH: Path = Field(default=None)
    QUERY_SIZE: int = Field(default=50_000)
    MATCH_BY: str = Field(default="text")
    TOP_K: int = Field(default=10)

    @computed_field
    @property
    def col_embedding(self) -> str:
        dict_ = {"image": "image_embed"}
        return dict_.get(self.MATCH_BY, "text_embed")

    @computed_field
    @property
    def path_array_db(self) -> Path:
        return self.ROOT_PATH / f"db_array"

    @computed_field
    @property
    def path_ds_db(self) -> Path:
        return self.ROOT_PATH / f"db_ds"

    @computed_field
    @property
    def path_ds_q(self) -> Path:
        return self.ROOT_PATH / f"q_ds"

    @computed_field
    @property
    def path_ds_inner(self) -> Path:
        return self.ROOT_PATH / f"ds"

    @computed_field
    @property
    def path_array_inner(self) -> Path:
        return self.ROOT_PATH / f"array"

    @computed_field
    @property
    def path_index(self) -> Path:
        return self.ROOT_PATH / f"index"

    @computed_field
    @property
    def path_result_query_score(self) -> Path:
        path_result = self.ROOT_PATH / f"result"
        make_dir(path_result)
        return path_result

    @computed_field
    @property
    def path_result_final(self) -> Path:
        path_result = self.ROOT_PATH / f"result_match_{self.MATCH_BY}"
        make_dir(path_result)
        return path_result


class BuildIndexAndQuery:
    def __init__(
        self,
        file_export_name: str,
        config: ConfigQuery,
        inner: bool = False,
        explode: bool = True,
    ):
        self.TOP_K = config.TOP_K
        self.QUERY_SIZE = config.QUERY_SIZE
        self.col_embedding = config.col_embedding
        self.inner = inner
        self.explode = explode
        self.sort_key_ds = lambda x: int(x.stem)
        self.sort_key_result = lambda x: int(x.stem.split("_")[1])

        # index
        self.path_index = config.path_index
        self.file_index = self.path_index / f"ip.index"
        self.file_index_json = str(self.path_index / f"index.json")

        # array
        self.path_array_db = config.path_array_inner if inner else config.path_array_db

        # ds
        self.dataset_dict = {
            "db_ds_path": config.path_ds_db,
            "q_ds_path": config.path_ds_q,
            "inner_ds_path": config.path_ds_inner,
        }

        # result
        self.path_result_query_score = config.path_result_query_score
        self.path_result_final = config.path_result_final
        self.file_export_final = self.path_result_final / f"{file_export_name}.parquet"

    def build(self):
        # Build index
        start = perf_counter()
        if not self.file_index.exists():
            print(f"[BuildIndex] Start")
            try:
                build_index(
                    str(self.path_array_db),
                    index_path=str(self.file_index),
                    index_infos_path=self.file_index_json,
                    save_on_disk=True,
                    metric_type="ip",
                    verbose=30,
                )
                print(f"[BuildIndex] Time finished: {perf_counter() - start:,.2f}s")
            except TypeError as e:
                print(f"[BuildIndex] Error: {e}")
                return None
        else:
            print(f"[BuildIndex] Index is existed")

    def load_dataset(self):
        dataset = {}
        if self.inner:
            for i in ["db", "q"]:
                dataset[i] = concatenate_datasets(
                    [
                        load_from_disk(str(f))
                        for f in sorted(
                            self.dataset_dict[f"inner_ds_path"].glob("*"),
                            key=self.sort_key_ds,
                        )
                    ]
                )

                for c in dataset[i].column_names:
                    if c != self.col_embedding:
                        dataset[i] = dataset[i].rename_column(c, f"{i}_{c}")

        else:
            for i in ["db", "q"]:
                dataset[i] = concatenate_datasets(
                    [
                        load_from_disk(str(f))
                        for f in sorted(
                            self.dataset_dict[f"{i}_ds_path"].glob("*"),
                            key=self.sort_key_ds,
                        )
                    ]
                )

        # Add index
        dataset["db"].load_faiss_index(self.col_embedding, self.file_index)
        return dataset["db"], dataset["q"]

    def query(self):
        # Load dataset
        dataset_db, dataset_q = self.load_dataset()

        # Batch query
        run = create_batch_index(len(dataset_q), self.QUERY_SIZE)
        num_batches = len(run)
        for i, val in run.items():
            # init
            file_name_result = self.path_result_query_score / f"result_{i}.parquet"
            file_name_score = self.path_result_query_score / f"score_{i}.parquet"
            if file_name_result.exists():
                continue

            # query
            start_idx, end_idx = val[0], val[-1]
            if start_idx == end_idx:  # prevent sample size is 1
                end_idx = None

            start_batch = perf_counter()
            score, result = dataset_db.get_nearest_examples_batch(
                self.col_embedding,
                np.asarray(dataset_q[start_idx:end_idx][self.col_embedding]),
                k=self.TOP_K,
            )
            # export
            for arr in result:
                del arr[self.col_embedding]  # prevent memory leaks
            df_result = pl.DataFrame(result)
            df_result.write_parquet(file_name_result)

            # track errors
            if df_result.shape[0] == 0:
                print(f"[red]No matches found for {i}[/]")
                continue

            dict_ = {
                f"score_{self.col_embedding}": [list(np.round(arr, 6)) for arr in score]
            }
            df_score = pl.DataFrame(dict_)
            df_score.write_parquet(file_name_score)

            # log
            end = perf_counter() - start_batch
            print(
                f"[Query] Batch {i}/{num_batches - 1} match result shape: {df_result.shape} {end:,.2f}s"
            )
            del score, result, df_score, df_result

        # Post process
        dataset_q = dataset_q.remove_columns(
            self.col_embedding
        )  # prevent polars issues
        del dataset_db

        df_score = pl.concat(
            [
                pl.read_parquet(f)
                for f in sorted(
                    self.path_result_query_score.glob("score*.parquet"),
                    key=self.sort_key_result,
                )
            ]
        )
        df_result = pl.concat(
            [
                pl.read_parquet(f)
                for f in sorted(
                    self.path_result_query_score.glob("result*.parquet"),
                    key=self.sort_key_result,
                )
            ]
        )

        df_match = pl.concat(
            [dataset_q.to_polars(), df_result, df_score], how="horizontal"
        )
        if self.explode:
            df_match = data_explode_list(df_match)

        df_match.write_parquet(self.file_export_final)



import polars as pl
import numpy as np
from FlagEmbedding import BGEM3FlagModel
from core_pro.ultilities import make_sync_folder


def load():
    path = make_sync_folder('Item_Matching_Test')
    file = path / 'Description_Images.parquet'

    col = [
        'item_id',
        'level1_global_be_category',
        'item_name',
        'image_url',
    ]
    df = (
        pl.read_parquet(file)
        .select(col)
        .unique(subset=col)
        .with_row_index('id')
    )

    print(f'Data Shape: {df.shape}')
    return df, col, path


def embed(df, path):
    item = df['q_item_name_clean'].to_list()
    file_embed = path / 'embed.npy'
    if not file_embed.exists():
        model = BGEM3FlagModel('BAAI/bge-m3', use_fp16=True)
        embeddings = model.encode(
            item,
            batch_size=512,
            max_length=80,
            return_dense=True,
            return_sparse=False,
            return_colbert_vecs=False
        )['dense_vecs']
        np.save(file_embed, embeddings)
    else:
        embeddings = np.load(file_embed)

    df = df.with_columns(pl.Series(values=embeddings, name='vector'))
    return embeddings, file_embed, df



from tqdm.auto import tqdm
from time import perf_counter
from pymilvus import MilvusClient, Collection
import polars as pl
import numpy as np
from core_pro.ultilities import make_sync_folder, create_batch_index


# path
path = make_sync_folder('dataset/item_matching')
file = path / 'data_sample_FMCG_clean.parquet'

df = pl.read_parquet(file)
print(df.shape)
print(df.columns)
embeddings = np.load(path / "embeddings.npy")

drop_col = ['cluster', 'description', 'images', 'image_url']
df = (
    df.with_columns(pl.Series('embeddings', embeddings))
    .drop(drop_col)
)

client = MilvusClient("milvus_demo.db")
collection_name = 'demo_collection'
if client.has_collection(collection_name=collection_name):
    client.drop_collection(collection_name=collection_name)

client.create_collection(
    collection_name=collection_name,
    dimension=embeddings.shape[1],
    primary_field_name='item_id',
    metric_type="COSINE",
    enable_dynamic_field=True,
    vector_field_name="embeddings",
)

start = perf_counter()
QUERY_SIZE = 1000
run = create_batch_index(len(embeddings), QUERY_SIZE)
data = df.to_dicts()
for i, val in tqdm(run.items()):
    start_idx, end_idx = val[0], val[-1]
    batch = data[start_idx:end_idx]
    res = client.insert(collection_name=collection_name, data=batch)

end = perf_counter() - start
print(f'Total Index Time: {end:,.0f}s')

start = perf_counter()
client.load_collection(collection_name=collection_name)
search_params = {
    "metric_type": "COSINE",
    "params": {
        "ef": 64  # Higher values give more accurate results but slower performance
    }
}

total_sample = len(data)
batch_size = 100
run = create_batch_index(len(embeddings), QUERY_SIZE)
for i, val in tqdm(run.items()):
    start_idx, end_idx = val[0], val[-1]
    batch = embeddings[start_idx:end_idx].tolist()
    res = client.search(
        collection_name=collection_name,
        data=batch,
        limit=10,
        search_params=search_params,
        output_fields=[i for i in df.columns if i != "embeddings"],
    )

end = perf_counter() - start
print(f'Total Query Time: {end:,.0f}s')


import polars as pl
import duckdb
from core_eda import TextEDA
from rich import print


class PipelineText:
    def __init__(self, mode: str = ""):
        self.mode = mode
        print(f"[Text Cleaning] {mode}")

    def run(self, data, key_col: list = None):
        # load data
        query = f"""select * from data"""
        df = duckdb.sql(query).pl()
        print(f"-> Base Data {self.mode}: {df.shape}")

        df = df.pipe(TextEDA.clean_text_pipeline_polars).drop_nulls(subset=key_col)
        if self.mode != "":
            df = df.select(pl.all().name.prefix(f"{self.mode}_"))
        print(f"-> Cleaned Data {self.mode}: {df.shape}")
        return df
